import concurrent.futures
import json
from typing import Optional, List, Tuple
from lxml import etree
from urllib.error import URLError
from urllib.parse import urljoin, urlencode, parse_qs
from urllib.request import urlopen
from .cache import Cache


class WikiData:

    class WikiDataException(Exception):

        def __init__(self, message: str, source_exception: Exception = None):
            self.source_exception = source_exception
            super().__init__(message)

    class __FetchException(WikiDataException):

        def __init__(
                self,
                message: str,
                page_id: str = None,
                source_exception: Exception = None):
            if page_id:
                super().__init__(
                    'Could not fetch the page: {}\n{}'.format(
                        page_id, message),
                    source_exception)
            else:
                super().__init__(message, source_exception)

    def __init__(
            self,
            cache: Cache = None,
            wikipedia_url: str = 'https://en.wikipedia.org/'):
        self.__cache = cache
        self.__wikipedia_url = wikipedia_url
        self.__timeout: int = 30
        self.__max_worker_count: int = 8
        self.__tables: list[etree.Element] = []
        self.__lists: list[etree.Element] = []
        if not self.__cache:
            self.__cache = Cache()
            self.__cache.setup()

    def __get_text(self, node: etree.Element) -> Optional[str]:
        if node is None:
            return None
        if not node.text:
            if len(node) > 0:
                return self.__get_text(node[0])
            return None
        return node.text

    def __is_sublist(self, node: etree.Element) -> bool:
        for child_node in node:
            if child_node.tag in ['ul', 'ol', 'dl']:
                return True
        return False

    def __get_url(
            self,
            page_id: str,
            section_id: int = None,
            wikipedia_url: str = None) -> str:
        if not wikipedia_url:
            wikipedia_url = self.__wikipedia_url
        url_base = urljoin(wikipedia_url, '/w/api.php')
        url_params = {
            'action': 'parse',
            'page': page_id,
            'format': 'json'
        }
        if section_id != None:
            url_params['section'] = section_id
            url_params['prop'] = 'text'
            url_params['disabletoc'] = '1'
            url_params['disableeditsection'] = '1'
        else:
            url_params['prop'] = 'sections'
        return url_base + '?' + urlencode(url_params)

    def __fetch_url(self, url: str) -> str:
        page_id = None
        if 'page' in parse_qs(url):
            page_id = parse_qs(url)['page'][0]
        try:
            response = urlopen(url, timeout=self.__timeout)
            data = json.loads(response.read())
        except URLError as e:
            raise WikiData.__FetchException(
                'Could not fetch the URL: {}'.format(url), page_id, e)
        except json.JSONDecodeError as e:
            raise WikiData.__FetchException(
                'The response is not a valid JSON text.', page_id, e)
        if 'error' in data and 'info' in data['error']:
            raise WikiData.__FetchException(
                'Wikipedia API: {}'.format(data['error']['info']), page_id)
        return data

    def __fetch_section_list(
            self,
            page_id: str,
            wikipedia_url: str = None) -> List[Tuple[int, str]]:
        url = self.__get_url(page_id, None, wikipedia_url)
        data = self.__cache.read(url)
        if not data:
            data = self.__fetch_url(url)
            if 'parse' not in data or 'sections' not in data['parse']:
                raise WikiData.__FetchException(
                    'The response has an unexpected format.', page_id)
            data = data['parse']['sections']
            data = [(int(x['index']), x['line']) for x in data if x['index']]
            data.insert(0, (0, ''))
            self.__cache.write(url, json.dumps(data))
        else:
            data = json.loads(data)
        return data

    def __fetch_section(
            self,
            page_id: str,
            section_id: int,
            wikipedia_url: str = None) -> str:
        url = self.__get_url(page_id, section_id, wikipedia_url)
        data = self.__cache.read(url)
        if not data:
            data = self.__fetch_url(url)
            if 'parse' not in data or 'text' not in data['parse'] or \
                    '*' not in data['parse']['text']:
                raise WikiData.__FetchException(
                    'The response has an unexpected format.', page_id)
            data = data['parse']['text']['*']
            self.__cache.write(url, data)
        return data

    def __process_section(self, data: str):
        root = etree.HTML(data)
        for table in root.iter('tbody'):
            self.__tables.append(table)
        for list_tag in ['ul', 'ol', 'dl']:
            for list_node in root.iter(list_tag):
                self.__lists.append(list_node)

    def fetch(
            self,
            page_id: str,
            excluded_sections: List[str] = [],
            wikipedia_url: str = None):
        self.__tables.clear()
        self.__lists.clear()
        section_list = self.__fetch_section_list(page_id, wikipedia_url)
        with concurrent.futures.ThreadPoolExecutor(
                max_workers=self.__max_worker_count) as executor:
            future_list = []
            for section in section_list:
                if section[1] in excluded_sections:
                    continue
                future = executor.submit(
                    self.__fetch_section, page_id, section[0], wikipedia_url)
                future_list.append(future)
            for future in concurrent.futures.as_completed(future_list):
                data = future.result()
                self.__process_section(data)

    def get_table_count(self) -> int:
        return len(self.__tables)

    def get_list_count(self) -> int:
        return len(self.__lists)

    def get_table_headers(self, table_index: int) -> List[str]:
        headers: list[str] = []
        rows = iter(self.__tables[table_index])
        if not rows:
            return None
        for row in rows:
            for cell in row:
                text = self.__get_text(cell)
                if text:
                    headers.append(text.strip())
                else:
                    headers.append(None)
            break
        return headers

    def get_table_values_by_column(
            self, table_index: int, column_index: int) -> List[str]:
        values: list[str] = []
        rows = iter(self.__tables[table_index])
        for row in rows:
            if column_index >= len(row):
                continue
            value = self.__get_text(row[column_index])
            if value:
                values.append(value)
        return values[1:]

    def get_table_values_by_header(
            self, table_index: int, header: str) -> List[str]:
        headers = self.get_table_headers(table_index)
        if not headers:
            return []
        if header in headers:
            return self.get_table_values_by_column(
                table_index, headers.index(header))
        elif header.strip() in headers:
            return self.get_table_values_by_column(
                table_index, headers.index(header.stip()))
        return []

    def get_list_values(self, list_index: int) -> List[str]:
        values: list[str] = []
        rows = iter(self.__lists[list_index])
        for row in rows:
            if row.tag not in ['li', 'dt']:
                continue
            if self.__is_sublist(row):
                continue
            value = self.__get_text(row)
            if value:
                values.append(value)
        return values



import hashlib
import yaml
from distutils.dir_util import copy_tree
from distutils.errors import DistutilsFileError, DistutilsInternalError
from os import listdir, makedirs, path, remove
from typing import Any, Optional, Union, List


class Config:

    TRANSFORM_CASE_OPTIONS = ['keep', 'lower', 'upper']

    class ConfigException(Exception):

        def __init__(
                self, message: str,
                source_file: str = None,
                source_exception: Exception = None):
            self.source_exception = source_exception
            if source_file:
                super().__init__(
                    'Could not load the configuration file: {}\n{}'.format(
                        source_file, message))
            else:
                super().__init__(message)

    class __MissingParameterException(ConfigException):

        def __init__(self, parameter_name: str):
            super().__init__(
                'The parameter is missing: {}'.format(
                    parameter_name))

    class __InvalidParameterTypeException(ConfigException):

        def __init__(self, parameter_name: str, parameter_type: type):
            super().__init__(
                'The type is not allowed for the \'{}\' parameter: {}'.format(
                    parameter_name, parameter_type.__name__))

    class __InvalidParameterValueException(ConfigException):

        def __init__(self, parameter_name: str, parameter_value: object):
            super().__init__(
                'The value is not allowed for the \'{}\' parameter: {}'.format(
                    parameter_name, parameter_value))

    def __init__(self, base_path: str = 'config/'):
        self.__base_path = base_path
        self.__wikipedia_url = None
        self.__excluded_sections = None
        self.__profiles = {}

    def __load_file(self, file_path: str) -> dict:
        try:
            file = open(file_path, 'r')
            data = yaml.load(file, Loader=yaml.SafeLoader)
            file.close()
            if not isinstance(data, dict):
                raise Config.ConfigException(
                    'The file syntax is invalid.', file_path)
            return data
        except Exception as e:
            if isinstance(e, Config.ConfigException):
                raise e
            else:
                raise Config.ConfigException(
                    'The file does not exist or its syntax is invalid.',
                    file_path,
                    e)

    def __load_parameter(
            self,
            data: dict,
            key: str,
            optional: bool = False,
            types: Union[type, List[type]] = str,
            name: str = None) -> Any:
        display_name = name
        if not display_name:
            display_name = key
        parameter = data.get(key)
        if parameter or parameter == 0:
            if not isinstance(parameter, types):
                raise Config.__InvalidParameterTypeException(
                    display_name, type(parameter))
        elif not optional:
            raise Config.__MissingParameterException(key)
        return parameter

    def __load_profile(self, name: str, data: dict) -> dict:
        if not name:
            raise self.__MissingParameterException('profile/name')
        if '{' in name or '}' in name:
            raise self.__InvalidParameterValueException('profile/name', name)
        if self.get_profile(name):
            raise self.ConfigException(
                'The profile is already defined: {}'.format(name))
        profile = {}
        profile['name'] = name
        profile['pattern'] = '{{{}}}'.format(name)
        if data:
            profile['transform_case'] = self.__load_parameter(
                data,
                'transform_case',
                True,
                name='profile/transform_case')
            profile['transform_space'] = self.__load_parameter(
                data, 'transform_space',
                True,
                (bool, str),
                'profile/transform_space')
            profile['transform_unidecode'] = self.__load_parameter(
                data, 'transform_unidecode',
                True,
                bool,
                'profile/transform_unidecode')
            profile['validation_pattern'] = self.__load_parameter(
                data, 'validation_pattern',
                True,
                name='profile/validation_pattern')
        else:
            profile['transform_case'] = None
            profile['transform_space'] = None
            profile['transform_unidecode'] = None
            profile['validation_pattern'] = None
        if profile['transform_case']:
            if profile['transform_case'] not in self.TRANSFORM_CASE_OPTIONS:
                raise self.__InvalidParameterValueException(
                    'profile/transform_case', profile['transform_case'])
        else:
            profile['transform_case'] = self.TRANSFORM_CASE_OPTIONS[0]
        if not profile['transform_space']:
            profile['transform_space'] = False
        if not profile['transform_unidecode']:
            profile['transform_unidecode'] = False
        if not profile['validation_pattern']:
            profile['validation_pattern'] = '.*'
        return profile

    def __load_main(self, data: dict):
        self.__wikipedia_url = self.__load_parameter(data, 'wikipedia_url')
        self.__excluded_sections = self.__load_parameter(
            data, 'excluded_sections', True, types=list)
        profiles_raw = self.__load_parameter(data, 'profile', types=list)
        for profile_data in profiles_raw:
            profile_name = self.__load_parameter(
                profile_data, 'name', name='profile/name')
            profile_pattern = self.__load_parameter(
                profile_data, 'pattern', name='profile/pattern')
            profile = self.__load_profile(profile_name, profile_data)
            profile['pattern'] = profile_pattern
            self.__profiles[profile['name']] = profile
        if self.__excluded_sections:
            for value in self.__excluded_sections:
                if not isinstance(value, str):
                    raise self.__InvalidParameterTypeException(
                        'excluded_sections/value', type(value))
        else:
            self.__excluded_sections = []
        if not self.get_profile('main'):
            raise self.ConfigException('The \'main\' profile is not defined.')

    def __load_code_name_list(self, name: str, data: dict):
        profile_data = self.__load_parameter(data, 'profile', True, dict)
        profile = self.__load_profile(name, profile_data)
        code_name_list = {}
        code_name_list['wikipedia_url'] = self.__load_parameter(
            data, 'wikipedia_url', True)
        code_name_list['excluded_sections'] = self.__load_parameter(
            data, 'excluded_sections', True, types=list)
        code_name_list['pages'] = self.__load_parameter(
            data, 'pages', types=list)
        code_name_list['sources'] = self.__load_parameter(
            data, 'sources', types=dict)
        for value in code_name_list['pages']:
            if not isinstance(value, str):
                raise self.__InvalidParameterTypeException(
                    'pages/value', type(value))
        if not code_name_list['wikipedia_url']:
            code_name_list['wikipedia_url'] = False
        if not code_name_list['excluded_sections']:
            code_name_list['excluded_sections'] = False
        if 'lists' in code_name_list['sources']:
            if not isinstance(code_name_list['sources']['lists'], bool):
                raise self.__InvalidParameterTypeException(
                    'sources/lists', type(code_name_list['sources']['lists']))
        else:
            code_name_list['sources']['lists'] = False
        if 'tables' in code_name_list['sources']:
            if not isinstance(code_name_list['sources']['tables'], list):
                raise self.__InvalidParameterTypeException(
                    'sources/tables', type(code_name_list['sources']['tables']))
            for value in code_name_list['sources']['tables']:
                if not isinstance(value, str):
                    raise self.__InvalidParameterTypeException(
                        'sources/tables/value', type(value))
        else:
            code_name_list['sources']['tables'] = []
        profile['code_name_list'] = code_name_list
        self.__profiles[profile['name']] = profile

    def generate(self):
        try:
            makedirs(self.__base_path, exist_ok=True)
        except OSError as e:
            raise Config.ConfigException(
                'Could not create the configuration directory: {}'.format(
                    self.__base_path),
                None,
                e)
        for file_name in listdir(self.__base_path):
            file_path = path.join(self.__base_path, file_name)
            if path.isfile(file_path):
                file_extension = path.splitext(file_name)[1]
                if file_extension.lower() != '.yaml':
                    continue
                try:
                    if path.isfile(file_path):
                        remove(file_path)
                except Exception as e:
                    raise Config.ConfigException(
                        'Could not delete the file: {}'.format(file_path), e)
        script_directory_path = path.dirname(path.realpath(__file__))
        default_config_path = path.join(
            script_directory_path, 'defaults', 'config')
        try:
            copy_tree(default_config_path, self.__base_path)
        except (DistutilsFileError, DistutilsInternalError) as e:
            raise Config.ConfigException(
                'Could not copy the default configuration: {} -> {}'.format(
                    default_config_path, self.__base_path),
                None,
                e)

    def load(self):
        if not path.isdir(self.__base_path):
            raise Config.ConfigException(
                'The configuration directory does not exist: {}'.format(
                    self.__base_path))
        for file_name in listdir(self.__base_path):
            file_path = path.join(self.__base_path, file_name)
            if path.isfile(file_path):
                file_base_name = path.splitext(file_name)[0]
                file_extension = path.splitext(file_name)[1]
                if file_extension.lower() != '.yaml':
                    continue
                try:
                    data = self.__load_file(file_path)
                    if file_base_name.lower() == 'main':
                        self.__load_main(data)
                    else:
                        self.__load_code_name_list(
                            file_base_name.lower(), data)
                except (Config.__MissingParameterException,
                        Config.__InvalidParameterTypeException,
                        Config.__InvalidParameterValueException) as e:
                    raise Config.ConfigException(
                        str(e), file_path, e.source_exception)
        if not self.get_profile('main'):
            raise self.ConfigException('The \'main\' profile is not defined.')

    def get_version(self) -> str:
        hash = hashlib.sha1()
        hash.update(str(self.__dict__).encode('utf-8'))
        return hash.hexdigest()

    def get_wikipedia_url(self) -> Optional[str]:
        return self.__wikipedia_url

    def get_excluded_sections(self) -> Optional[List[str]]:
        return self.__excluded_sections

    def get_profile(self, name: str) -> dict:
        return self.__profiles.get(name.lower())

    def get_profile_name_list(self) -> List[str]:
        return list(self.__profiles.keys())



import argparse
import sys
from appdirs import user_cache_dir, user_config_dir
from colorama import Fore, init as colorama_init
from os import path
from .cache import Cache
from .generator import Generator
from .config import Config

CONFIG_PATH_FLAG = '--config-path'
CACHE_PATH_FLAG = '--cache-path'
PROFILE_FLAG = '--profile'
PROFILE_FLAG_SHORT = '-p'
COUNT_FLAG = '--count'
COUNT_FLAG_SHORT = '-c'
ATTEMPT_COUNT_FLAG = '--attempt-count'
SORT_FLAG = '--sort'
SORT_FLAG_SHORT = '-s'
LIST_ALL_FLAG = '--list-all'
LIST_PROFILES_FLAG = '--list-profiles'
GENERATE_CONFIG_FLAG = '--generate-config'
CLEAR_CACHE_FLAG = '--clear-cache'
QUIET_FLAG = '--quiet'
QUIET_FLAG_SHORT = '-q'

APPLICATION_MESSAGE = 'Generate code names using lists and tables from ' \
    'Wikipedia articles.'
CONFIG_PATH_FLAG_MESSAGE = 'set a path to the configuration directory'
CACHE_PATH_FLAG_MESSAGE = 'set a path to the cache directory'
PROFILE_FLAG_MESSAGE = 'set a profile name'
COUNT_FLAG_MESSAGE = 'set a length of the generated list of code names'
ATTEMPT_COUNT_FLAG_MESSAGE = 'set a maximum number of attempts to generate ' \
    'a valid code name'
SORT_FLAG_MESSAGE = 'sort the generated list of code names'
LIST_ALL_FLAG_MESSAGE = 'list all code names for the profile (must be a list ' \
    'of code names)'
LIST_PROFILES_FLAG_MESSAGE = 'list all available profiles'
GENERATE_CONFIG_FLAG_MESSAGE = 'generate a default configuration'
CLEAR_CACHE_FLAG_MESSAGE = 'clear the cache'
QUIET_FLAG_MESSAGE = 'do not print additional messages (useful in scripts)'

APP_DIRECTORY_NAME = 'wikicodename'
CONFIG_VERSION_CACHE_KEY = 'config_version'

INITIAL_CONFIG_GENERATION_FAILED_MESSAGE = '{}Failed to generate the initial ' \
    'configuration.{}'.format(Fore.RED, Fore.RESET)
CONFIG_GENERATED_MESSAGE = '{}The configuration has been generated.{}'.format(
    Fore.GREEN, Fore.RESET)
CONFIG_SHOULD_BE_GENERATED_OR_FIXED_MESSAGE = '{}Please run the application ' \
    'with the {} flag to generate a valid configuration or fix it manually.' \
    '{}'.format(Fore.YELLOW, GENERATE_CONFIG_FLAG, Fore.RESET)
CONFIG_CHANGED_MESSAGE = '{}The configuration has changed. Please run the ' \
    'application with the {} flag to clear the cache.{}'.format(
        Fore.YELLOW, CLEAR_CACHE_FLAG, Fore.RESET)
CACHE_CLEARED_MESSAGE = '{}The cache has been cleared.{}'.format(
    Fore.GREEN, Fore.RESET)


def get_default_config_path():
    return user_config_dir(APP_DIRECTORY_NAME, False)


def get_default_cache_path():
    return user_cache_dir(APP_DIRECTORY_NAME, False)


def parse_args():
    arg_parser = argparse.ArgumentParser(
        description=APPLICATION_MESSAGE, add_help=False)
    arg_parser.add_argument(
        '--help', '-h', action='help', help='show this help message')
    arg_parser.add_argument(
        CONFIG_PATH_FLAG,
        type=str,
        nargs=1,
        default=get_default_config_path(),
        help=CONFIG_PATH_FLAG_MESSAGE)
    arg_parser.add_argument(
        CACHE_PATH_FLAG,
        type=str,
        nargs=1,
        default=get_default_cache_path(),
        help=CACHE_PATH_FLAG_MESSAGE)
    arg_parser.add_argument(
        PROFILE_FLAG,
        PROFILE_FLAG_SHORT,
        type=str,
        nargs=1,
        default='main',
        help=PROFILE_FLAG_MESSAGE)
    arg_parser.add_argument(
        COUNT_FLAG,
        COUNT_FLAG_SHORT,
        type=int,
        nargs=1,
        default=10,
        help=COUNT_FLAG_MESSAGE)
    arg_parser.add_argument(
        ATTEMPT_COUNT_FLAG,
        type=int,
        nargs=1,
        default=64,
        help=ATTEMPT_COUNT_FLAG_MESSAGE)
    arg_parser.add_argument(
        SORT_FLAG,
        SORT_FLAG_SHORT,
        action='store_const',
        const=True,
        default=False,
        help=SORT_FLAG_MESSAGE)
    arg_parser.add_argument(
        LIST_ALL_FLAG,
        action='store_const',
        const=True,
        default=False,
        help=LIST_ALL_FLAG_MESSAGE)
    arg_parser.add_argument(
        LIST_PROFILES_FLAG,
        action='store_const',
        const=True,
        default=False,
        help=LIST_PROFILES_FLAG_MESSAGE)
    arg_parser.add_argument(
        GENERATE_CONFIG_FLAG,
        action='store_const',
        const=True,
        default=False,
        help=GENERATE_CONFIG_FLAG_MESSAGE)
    arg_parser.add_argument(
        CLEAR_CACHE_FLAG,
        action='store_const',
        const=True,
        default=False,
        help=CLEAR_CACHE_FLAG_MESSAGE)
    arg_parser.add_argument(
        QUIET_FLAG,
        QUIET_FLAG_SHORT,
        action='store_const',
        const=True,
        default=False,
        help=QUIET_FLAG_MESSAGE)
    return arg_parser.parse_args()


def get_arg(value):
    if isinstance(value, list):
        return value[0]
    return value


def print_exception(exception: Exception):
    print('\r{}{}{}'.format(Fore.RED, exception, Fore.RESET), file=sys.stderr)
    if hasattr(exception, 'source_exception'):
        if exception.source_exception:
            print(exception.source_exception, file=sys.stderr)


def main():
    colorama_init()
    args = parse_args()
    config_path = get_arg(args.config_path)
    if config_path == get_default_config_path():
        if not path.isdir(config_path):
            try:
                initial_config = Config(config_path)
                initial_config.generate()
            except Config.ConfigException as e:
                print(INITIAL_CONFIG_GENERATION_FAILED_MESSAGE)
                print_exception(e)
                return 1
    try:
        config = Config(config_path)
        if args.generate_config:
            config.generate()
            if not args.quiet:
                print(CONFIG_GENERATED_MESSAGE)
            return 0
        config.load()
        if args.list_profiles:
            profile_name_list = config.get_profile_name_list()
            profile_name_list.sort()
            for profile_name in profile_name_list:
                print(profile_name)
            return 0
    except Config.ConfigException as e:
        print(CONFIG_SHOULD_BE_GENERATED_OR_FIXED_MESSAGE)
        print_exception(e)
        return 2
    try:
        cache = Cache(get_arg(args.cache_path))
        cache.setup()
        if args.clear_cache:
            cache.clear()
            if not args.quiet:
                print(CACHE_CLEARED_MESSAGE)
            return 0
        if not args.quiet:
            config_version = cache.read(CONFIG_VERSION_CACHE_KEY)
            if config_version and config_version != config.get_version():
                print(CONFIG_CHANGED_MESSAGE)
        cache.write(CONFIG_VERSION_CACHE_KEY, config.get_version())
    except Cache.CacheException as e:
        print_exception(e)
        return 3
    try:
        generator = Generator(
            config,
            cache,
            get_arg(args.attempt_count),
            get_arg(args.quiet))
        code_name_list = None
        if args.list_all:
            code_name_list = generator.generate_all(
                get_arg(args.profile))
        else:
            code_name_list = generator.generate(
                get_arg(args.profile), get_arg(args.count))
        if args.sort:
            code_name_list.sort()
        for code_name in code_name_list:
            print(code_name)
    except Generator.GeneratorException as e:
        print_exception(e)
        return 4
    return 0


if __name__ == '__main__':
    sys.exit(main())



import hashlib
from os import listdir, makedirs, path, remove
from typing import Optional


class Cache:

    class CacheException(Exception):
        def __init__(self, message: str, source_exception: Exception = None):
            self.source_exception = source_exception
            super().__init__(message)

    def __init__(self, base_path: str = 'cache/'):
        self.__base_path = base_path

    def __get_file_path(self, id: str) -> str:
        hash = hashlib.sha1()
        hash.update(id.encode('utf-8'))
        return path.join(self.__base_path, hash.hexdigest())

    def setup(self) -> None:
        try:
            makedirs(self.__base_path, exist_ok=True)
        except OSError as e:
            raise Cache.CacheException(
                'Could not create the cache directory: {}'.format(
                    self.__base_path),
                e)

    def read(self, id: str) -> Optional[str]:
        data = None
        file_path = self.__get_file_path(id)
        try:
            file = open(file_path, 'r')
            data = file.read()
            file.close()
        except FileNotFoundError:
            pass
        except OSError as e:
            raise Cache.CacheException(
                'Could not read from the file: {}'.format(file_path), e)
        return data

    def write(self, id: str, data: str):
        file_path = self.__get_file_path(id)
        try:
            file = open(file_path, 'w')
            file.write(data)
            file.close()
        except OSError as e:
            raise Cache.CacheException(
                'Could not write to the file: {}'.format(file_path), e)

    def clear(self):
        for file_name in listdir(self.__base_path):
            file_path = path.join(self.__base_path, file_name)
            try:
                if path.isfile(file_path):
                    remove(file_path)
            except Exception as e:
                raise Cache.CacheException(
                    'Could not delete the file: {}'.format(file_path), e)



import json
import random
import re
from colorama import Fore
from typing import List, Dict
from unidecode import unidecode
from .cache import Cache
from .config import Config
from .wiki_data import WikiData


class Generator:

    class GeneratorException(Exception):

        def __init__(
                self,
                message,
                profile_name=None,
                source_exception=None) -> None:
            self.source_exception = source_exception
            if profile_name:
                super().__init__(
                    'Could not generate a list of code names for the profile: '
                    '{}\n{}'.format(profile_name, message))
            else:
                super().__init__(message)

    def __init__(
            self,
            config: Config = None,
            cache: Cache = None,
            max_attempt_count: int = 64,
            quiet: bool = False) -> None:
        self.__config = config
        self.__cache = cache
        self.__max_attempt_count = max_attempt_count
        self.__quiet = quiet
        if not self.__config:
            self.__config = Config()
        if not self.__cache:
            self.__cache = Cache()
            self.__cache.setup()

    def __parse_pattern(self, pattern: str) -> Dict[str, List[str]]:
        start = 0
        end = 0
        sequence_start = -1
        format_pattern = ''
        profiles: list[str] = []
        while end >= 0 and end < len(pattern):
            if sequence_start < 0:
                end = pattern.find('{', start)
            else:
                end = pattern.find('}', start)
            if end >= 0:
                if sequence_start < 0:
                    format_pattern += pattern[start:(end+1)]
                    sequence_start = end + 1
                else:
                    format_pattern += '}'
                    profile = pattern[sequence_start:end]
                    profiles.append(profile)
                    sequence_start = -1
                start = end + 1
                end = start
        format_pattern += pattern[start:]
        return {'format_pattern': format_pattern, 'profiles': profiles}

    def __format_code_name(self, code_name, profile) -> str:
        transformed_code_name = code_name
        transform_case = profile['transform_case']
        transform_space = profile['transform_space']
        transform_unidecode = profile['transform_unidecode']
        validation_pattern = profile['validation_pattern']
        if transform_case == 'lower':
            transformed_code_name = transformed_code_name.lower()
        elif transform_case == 'upper':
            transformed_code_name = transformed_code_name.upper()
        if transform_space != False:
            transformed_code_name = [
                transform_space if x.isspace() else x
                for x in transformed_code_name]
            transformed_code_name = ''.join(transformed_code_name)
        if transform_unidecode != False:
            transformed_code_name = unidecode(transformed_code_name)
        validation_result = re.search(
            validation_pattern, transformed_code_name)
        if not validation_result:
            return None
        return validation_result.group(0)

    def __get_code_name_list(self, profile_name: str) -> None:
        cache_name = 'profile_' + profile_name
        cache_data = self.__cache.read(cache_name)
        if cache_data:
            return json.loads(cache_data)
        if not self.__quiet:
            print('\r{}Fetching data for the profile: {}{}'.format(
                Fore.YELLOW, profile_name, Fore.RESET),
                end='')
        code_name_list = []
        profile = self.__config.get_profile(profile_name)
        if not profile:
            raise self.GeneratorException(
                'The profile is not defined.', profile_name)
        if 'code_name_list' not in profile:
            raise self.GeneratorException(
                'The profile does not define a list of code names.',
                profile_name)
        pages = profile['code_name_list']['pages']
        sources = profile['code_name_list']['sources']
        wikipedia_url = self.__config.get_wikipedia_url()
        excluded_sections = self.__config.get_excluded_sections()
        if profile['code_name_list']['wikipedia_url']:
            wikipedia_url = profile['code_name_list']['wikipedia_url']
        if profile['code_name_list']['excluded_sections']:
            excluded_sections = profile['code_name_list']['excluded_sections']
        data = WikiData(self.__cache, wikipedia_url)
        for page in pages:
            if not self.__quiet:
                print(
                    '\r{}Fetching data for the profile: {} '
                    '(page {}/{}){}'.format(
                        Fore.YELLOW,
                        profile_name,
                        pages.index(page) + 1,
                        len(pages),
                        Fore.RESET),
                    end='')
            data.fetch(page, excluded_sections, wikipedia_url)
            for i in range(data.get_table_count()):
                for header in sources['tables']:
                    code_name_list += data.get_table_values_by_header(
                        i, header)
            for i in range(data.get_list_count()):
                if sources['lists']:
                    code_name_list += data.get_list_values(i)
            code_name_list = [
                self.__format_code_name(x, profile) for x in code_name_list]
            code_name_list = [x for x in code_name_list if x]
        cache_data = json.dumps(code_name_list)
        self.__cache.write(cache_name, cache_data)
        if not self.__quiet:
            print('\r{}Fetched data for the profile: {}{}{}'.format(
                Fore.GREEN, profile_name, ' ' * (9 + 2 * 4), Fore.RESET))
        return code_name_list

    def __get_code_name(self, profile_name: str) -> str:
        attempt_count = 0
        code_name = None
        profile = self.__config.get_profile(profile_name)
        if not profile:
            raise self.GeneratorException(
                'The profile is not defined.', profile_name)
        pattern_data = profile['pattern']
        pattern = self.__parse_pattern(pattern_data)
        format_pattern = pattern['format_pattern']
        subprofile_names = pattern['profiles']
        subprofile_count = len(pattern['profiles'])
        if subprofile_count > 1:
            while not code_name and attempt_count < self.__max_attempt_count:
                subprofile_code_names = []
                if profile_name in subprofile_names:
                    raise self.GeneratorException(
                        'The user defined pattern must not contain its '
                        'profile: {}'.format(
                            pattern_data),
                        profile_name)
                for subprofile_name in subprofile_names:
                    subprofile_code_names.append(
                        self.__get_code_name(subprofile_name))
                code_name = self.__format_code_name(
                    format_pattern.format(*subprofile_code_names), profile)
                attempt_count += 1
        elif subprofile_count == 1:
            subprofile_name = subprofile_names[0]
            if profile_name == subprofile_name:
                code_name_list = self.__get_code_name_list(subprofile_name)
                if len(code_name_list) == 0:
                    raise self.GeneratorException(
                        'No code name match the profile.',
                        profile_name)
                code_name_index = random.randrange(0, len(code_name_list))
                code_name = code_name_list[code_name_index]
            else:
                while not code_name and attempt_count < self.__max_attempt_count:
                    subprofile_code_name = self.__get_code_name(
                        subprofile_name)
                    code_name = self.__format_code_name(
                        format_pattern.format(subprofile_code_name), profile)
                    attempt_count += 1
        else:
            raise self.GeneratorException(
                'The pattern does not contain any profile: {}'.format(
                    profile['pattern']),
                profile_name)
        if not code_name:
            raise self.GeneratorException(
                'The maximum number of attempts has been reached.')
        return code_name

    def generate(self, profile_name: str, count: int) -> List[str]:
        attempt_count = 0
        code_name_list: list[str] = []
        try:
            while len(code_name_list) < count and \
                    attempt_count < self.__max_attempt_count:
                code_name = self.__get_code_name(profile_name)
                if code_name not in code_name_list:
                    code_name_list.append(code_name)
                attempt_count += 1
        except Cache.CacheException as e:
            raise self.GeneratorException(
                str(e), profile_name, e.source_exception)
        except WikiData.WikiDataException as e:
            raise self.GeneratorException(
                str(e), profile_name, e.source_exception)
        if len(code_name_list) < count:
            raise self.GeneratorException(
                'The maximum number of attempts has been reached.')
        return code_name_list

    def generate_all(self, profile_name: str) -> List[str]:
        try:
            return self.__get_code_name_list(profile_name)
        except Cache.CacheException as e:
            raise self.GeneratorException(
                str(e), profile_name, e.source_exception)
        except WikiData.WikiDataException as e:
            raise self.GeneratorException(
                str(e), profile_name, e.source_exception)



import os
from setuptools import setup


def read(file_name):
    return open(os.path.join(os.path.dirname(__file__), file_name)).read()


setup(
    name='wikicodename',
    version='1.0.0',
    author='Bartłomiej Zdrojewski',
    author_email='bartlomiej.zdrojewski@gmail.com',
    url='https://github.com/bartlomiej-zdrojewski/wikicodename',
    description='Generate code names using lists and tables from Wikipedia '
        'articles.',
    long_description_content_type="text/markdown",
    long_description=read('README.md'),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Utilities'
    ],
    keywords='code name codename generate generator wiki Wikipedia',
    license='MIT',
    packages=['wikicodename'],
    include_package_data=True,
    package_data={
        'wikicodename': ['defaults/config/*']
    },
    install_requires=[
        'appdirs>=1.4,<1.5',
        'colorama>=0.4,<0.5',
        'lxml>=4.7,<4.8',
        'PyYAML>=6.0,<6.1',
        'Unidecode>=1.3,<1.4'
    ],
    entry_points={
        'console_scripts': [
            'wikicodename = wikicodename.__main__:main'
        ]
    }
)
